import {
    Address,
    CheckoutParams,
    CheckoutSelectors,
    Consignment,
    Country,
    CustomerAddress,
    CustomerRequestOptions,
    FormField,
    RequestOptions,
    ShippingInitializeOptions,
    ShippingRequestOptions,
} from '@bigcommerce/checkout-sdk'
import { withFormik, FormikProps } from 'formik'
import { debounce, noop } from 'lodash'
import React, { PureComponent, ReactNode } from 'react'
import { lazy, object } from 'yup'
import { connect } from 'react-redux'

import B3Checkout from '../b3checkout'
import {
    getAddressFormFieldsValidationSchema,
    getTranslateAddressError,
    isEqualAddress,
    mapAddressFromFormValues,
    mapAddressToFormValues,
    AddressFormValues,
} from '../address'
import { getCustomFormFieldsValidationSchema } from '../formFields'
import { withLanguage, WithLanguageProps } from '../locale'
import { Fieldset, Form, FormContext } from '../ui/form'

import hasSelectedShippingOptions from './hasSelectedShippingOptions'
import BillingSameAsShippingField from './BillingSameAsShippingField'
import ShippingAddress from './ShippingAddress'
import { SHIPPING_ADDRESS_FIELDS } from './ShippingAddressFields'
import ShippingFormFooter from './ShippingFormFooter'

export interface SingleShippingFormProps {
    addresses: CustomerAddress[]
    cartHasChanged: boolean
    consignments: Consignment[]
    countries: Country[]
    countriesWithAutocomplete: string[]
    customerMessage: string
    googleMapsApiKey?: string
    isLoading: boolean
    isShippingStepPending: boolean
    isMultiShippingMode: boolean
    methodId?: string
    shippingAddress?: Address
    shouldShowSaveAddress?: boolean
    shouldValidateSafeInput: boolean
    shouldShowOrderComments: boolean
    deinitialize(options: ShippingRequestOptions): Promise<CheckoutSelectors>
    deleteConsignments(): Promise<Address | undefined>
    getFields(countryCode?: string): FormField[]
    initialize(options: ShippingInitializeOptions): Promise<CheckoutSelectors>
    onSubmit(values: SingleShippingFormValues): void
    onUnhandledError?(error: Error): void
    signOut(options?: CustomerRequestOptions): void
    updateAddress(
        address: Partial<Address>,
        options?: RequestOptions<CheckoutParams>
    ): Promise<CheckoutSelectors>
}

export interface SingleShippingFormValues {
    billingSameAsShipping: boolean
    shippingAddress?: AddressFormValues
    orderComment: string
}

interface SingleShippingFormState {
    isResettingAddress: boolean
    isUpdatingShippingData: boolean
    hasRequestedShippingOptions: boolean
}

interface B3Props {
    b3Company: any
    isB3User: boolean
}

export const SHIPPING_AUTOSAVE_DELAY = 1700

class SingleShippingForm extends PureComponent<
    B3Props &
    SingleShippingFormProps &
    WithLanguageProps &
    FormikProps<SingleShippingFormValues>
    > {
    static contextType = FormContext

    state: SingleShippingFormState = {
        isResettingAddress: false,
        isUpdatingShippingData: false,
        hasRequestedShippingOptions: false,
    }

    private debouncedUpdateAddress: any

    constructor(
        props: B3Props &
            SingleShippingFormProps &
            WithLanguageProps &
            FormikProps<SingleShippingFormValues>
    ) {
        super(props)

        const { updateAddress } = this.props

        this.debouncedUpdateAddress = debounce(
            async (address: Address, includeShippingOptions: boolean) => {
                try {
                    await updateAddress(address, {
                        params: {
                            include: {
                                'consignments.availableShippingOptions': includeShippingOptions,
                            },
                        },
                    })
                    if (includeShippingOptions) {
                        this.setState({ hasRequestedShippingOptions: true })
                    }
                } finally {
                    this.setState({ isUpdatingShippingData: false })
                }
            },
            SHIPPING_AUTOSAVE_DELAY
        )
    }

    get shouldShowBillingSameAsShipping() {
        const {
            isB3User,
            addresses,
            methodId,
            shippingAddress,
        } = this.props

        const PAYMENT_METHOD_VALID = ['amazon', 'amazonpay']
        const shouldShowBillingSameAsShipping = !PAYMENT_METHOD_VALID.some(
            method => method === methodId
        )
        
        if (!shippingAddress) return false
        if (!isB3User) return shouldShowBillingSameAsShipping

        const comparedKeys = Object.keys(shippingAddress)
        let comparedAddresses = addresses.map(address =>
            comparedKeys.reduce((result, key) => {
                ; (result as any)[key] = (address as any)[key]
                return result
            }, {})
        )

        const index = comparedAddresses.findIndex((customerAddress) => isEqualAddress(customerAddress as CustomerAddress, shippingAddress))
        const shouldB3ShowBillingSameAsShipping = ~index ? (addresses[index] as CustomerAddress & { isBilling: number }).isBilling : true

        return !!(shouldShowBillingSameAsShipping && shouldB3ShowBillingSameAsShipping)
    }

    render(): ReactNode {
        const {
            addresses,
            cartHasChanged,
            isLoading,
            onUnhandledError,
            methodId,
            shouldShowSaveAddress,
            countries,
            countriesWithAutocomplete,
            googleMapsApiKey,
            shippingAddress,
            consignments,
            shouldShowOrderComments,
            initialize,
            isValid,
            deinitialize,
            values: { shippingAddress: addressForm },
            isShippingStepPending,
        } = this.props

        const {
            isResettingAddress,
            isUpdatingShippingData,
            hasRequestedShippingOptions,
        } = this.state

        

        return (
            <Form autoComplete='on'>
                <Fieldset>
                    <ShippingAddress
                        addresses={addresses}
                        consignments={consignments}
                        countries={countries}
                        countriesWithAutocomplete={countriesWithAutocomplete}
                        deinitialize={deinitialize}
                        formFields={this.getFields(addressForm && addressForm.countryCode)}
                        googleMapsApiKey={googleMapsApiKey}
                        hasRequestedShippingOptions={hasRequestedShippingOptions}
                        initialize={initialize}
                        isLoading={isResettingAddress}
                        isShippingStepPending={isShippingStepPending}
                        methodId={methodId}
                        onAddressSelect={this.handleAddressSelect}
                        onFieldChange={this.handleFieldChange}
                        onUnhandledError={onUnhandledError}
                        onUseNewAddress={this.onUseNewAddress}
                        shippingAddress={shippingAddress}
                        shouldShowSaveAddress={shouldShowSaveAddress}
                    />
                    {this.shouldShowBillingSameAsShipping && B3Checkout.config.permissions['shipping.singleAddress.allowAsBilling'] && (
                        <div className='form-body'>
                            <BillingSameAsShippingField />
                        </div>
                    )}
                </Fieldset>

                <ShippingFormFooter
                    cartHasChanged={cartHasChanged}
                    isLoading={isLoading || isUpdatingShippingData}
                    isMultiShippingMode={false}
                    shouldDisableSubmit={this.shouldDisableSubmit()}
                    shouldShowOrderComments={shouldShowOrderComments}
                    shouldShowShippingOptions={isValid}
                />
            </Form>
        )
    }

    private shouldDisableSubmit: () => boolean = () => {
        const { isLoading, consignments, isValid } = this.props

        const { isUpdatingShippingData } = this.state

        if (!isValid) {
            return false
        }

        return (
            isLoading ||
            isUpdatingShippingData ||
            !hasSelectedShippingOptions(consignments)
        )
    }

    private handleFieldChange: (name: string) => void = async name => {
        const { setFieldValue } = this.props

        if (name === 'countryCode') {
            setFieldValue('shippingAddress.stateOrProvince', '')
            setFieldValue('shippingAddress.stateOrProvinceCode', '')
        }

        // Enqueue the following code to run after Formik has run validation
        await new Promise(resolve => setTimeout(resolve))

        const isShippingField = SHIPPING_ADDRESS_FIELDS.includes(name)

        const { hasRequestedShippingOptions } = this.state

        const { isValid } = this.props

        if (!isValid) {
            return
        }

        this.updateAddressWithFormData(
            isShippingField || !hasRequestedShippingOptions
        )
    }

    private updateAddressWithFormData(includeShippingOptions: boolean) {
        const {
            shippingAddress,
            values: { shippingAddress: addressForm },
        } = this.props

        const updatedShippingAddress =
            addressForm && mapAddressFromFormValues(addressForm)

        if (
            !updatedShippingAddress ||
            isEqualAddress(updatedShippingAddress, shippingAddress)
        ) {
            return
        }

        this.setState({ isUpdatingShippingData: true })
        this.debouncedUpdateAddress(updatedShippingAddress, includeShippingOptions)
    }

    private handleAddressSelect: (address: Address) => void = async address => {
        const {
            updateAddress,
            onUnhandledError = noop,
            values,
            setValues,
        } = this.props

        this.setState({ isResettingAddress: true })

        try {
            await updateAddress(address)

            setValues({
                ...values,
                shippingAddress: mapAddressToFormValues(
                    this.getFields(address.countryCode),
                    address
                ),
            })
        } catch (error) {
            onUnhandledError(error)
        } finally {
            this.setState({ isResettingAddress: false })
        }
    }

    private onUseNewAddress: () => void = async () => {
        const {
            deleteConsignments,
            onUnhandledError = noop,
            setValues,
            values,
            isB3User,
            b3Company,
        } = this.props

        this.setState({ isResettingAddress: true })

        try {
            const address = await deleteConsignments()
            setValues({
                ...values,
                shippingAddress: isB3User
                    ? b3Company.basicAddressInfo
                    : mapAddressToFormValues(
                        this.getFields(address && address.countryCode),
                        address
                    ),
            })
        } catch (e) {
            onUnhandledError(e)
        } finally {
            this.setState({ isResettingAddress: false })
        }
    }

    private getFields(countryCode: string | undefined): FormField[] {
        const { getFields } = this.props

        return getFields(countryCode)
    }
}

const mapStateToProps = (state: {
    global: { isB3User: boolean }
    company: any
}) => {
    const {
        global: { isB3User },
        company: b3Company,
    } = state

    return {
        isB3User,
        b3Company,
    }
}

export default connect(mapStateToProps)(
    withLanguage(
        withFormik<
            B3Props & SingleShippingFormProps & WithLanguageProps,
            SingleShippingFormValues
        >({
            handleSubmit: (
                values,
                { props: { onSubmit, isB3User, getFields, addresses } }
            ) => {
                const { shippingAddress } = values
                const formFields = getFields(
                    shippingAddress && shippingAddress.countryCode
                )
                
                let comparedAddresses = [...addresses]
                if (shippingAddress?.stateOrProvinceCode) comparedAddresses = comparedAddresses.map(comparedAddress => ({
                    ...comparedAddress,
                    stateOrProvince: '',
                  }))
                
                const isValid = B3Checkout.utils.util.validateAddress(mapAddressFromFormValues(shippingAddress as any), comparedAddresses, formFields)

                onSubmit({
                    ...values,
                    shippingAddress: {
                        ...values.shippingAddress,
                        addressId: isB3User && !isValid ? '' : undefined
                    } as unknown as AddressFormValues,
                })
            },
            mapPropsToValues: ({ getFields, shippingAddress, customerMessage }) => ({
                billingSameAsShipping: false,
                orderComment: customerMessage,
                shippingAddress: mapAddressToFormValues(
                    getFields(shippingAddress && shippingAddress.countryCode),
                    shippingAddress
                ),
            }),
            isInitialValid: ({ shippingAddress, getFields, language }) =>
                !!shippingAddress &&
                getAddressFormFieldsValidationSchema({
                    language,
                    formFields: getFields(shippingAddress.countryCode),
                }).isValidSync(shippingAddress),
            validationSchema: ({
                language,
                getFields,
                methodId,
                shouldValidateSafeInput,
            }: SingleShippingFormProps & WithLanguageProps) =>
                methodId
                    ? object({
                        shippingAddress: lazy<Partial<AddressFormValues>>(formValues =>
                            getCustomFormFieldsValidationSchema({
                                translate: getTranslateAddressError(language),
                                formFields: getFields(formValues && formValues.countryCode),
                            })
                        ),
                    })
                    : object({
                        shippingAddress: lazy<Partial<AddressFormValues>>(formValues =>
                            getAddressFormFieldsValidationSchema({
                                language,
                                shouldValidateSafeInput,
                                formFields: getFields(formValues && formValues.countryCode),
                            })
                        ),
                    }),
            enableReinitialize: false,
        })(SingleShippingForm)
    )
)
